#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

;;; one line to give the Gauche-archive and an idea of what it does.
;;; Copyright (C) 2018 aaauuue 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(add-load-path ".." :retrive #t)
(define-module util.archives
  (use srfi-1)
  (use srfi-35)
  (use srfi-60)
  (use util.match)
  (use file.util)
  (use gauche.uvector)
  (use gauche.vport)
  (use gauche.collection)
  (use gauche.mop.instance-pool)
  (use gauche.mop.validator)

  (use archive)

  (export
   <archive-base>
   <archive-error> 
   <archive-reader-error> <archive-writer-error> <archive-builder-error>
   <archive-entry-error> 
   archive-error?
   archive-error-message archive-error-debug-info
   archive-reader-error?
   archive-reader-error-message archive-reader-error-debug-info
   archive-writer-error?
   archive-writer-error-message archive-writer-error-debug-info
   archive-builder-error?
   archive-builder-error-message archive-builder-error-debug-info
   archive-entry-error?
   archive-entry-error-message archive-entry-error-debug-info
   archive-throw archive-try
   <archive-reader> <archive-writer> <archive-builder> <archive-entry>
   open-input-archive-from-name
   open-input-archive-from-memory
   close-input-archive
   make-archive-entry
   get-archive-entry-data
   flush-archive-entry-data
   create-entry-from-file
   read-file
   create-entry-from-memory
   get-entry
   read-data
   extract-all-files
   extract-files-to-disk
   extract-files-to-memory
   begin-iterator
   end-iterator
   call-with-iterator
   call-with-file-archive
   call-with-memory-archive
   create-archive
   open-output-archive
   close-output-archive
   append-entry
   remove-entry
   modify-entry
   open-build-archive-to-file
   open-build-archive-to-memory
   close-build-archive
   append-entry

   ;;; extra definition
   compression-level-validator boolean-validator 7zip-compress-validator 
   string-validator xar-validator xar-compress-validator
   zip-compression-validator option->string 
   define-archive-option-mixin define-archive-builder-class
   archive-file-builder->memory-builder
   archive-memory-builder->file-builder

   ;;; constant 
   ARCHIVE-FILTER-NONE     
   ARCHIVE-FILTER-GZIP     
   ARCHIVE-FILTER-BZIP2    
   ARCHIVE-FILTER-COMPRESS 
   ARCHIVE-FILTER-PROGRAM  
   ARCHIVE-FILTER-LZMA     
   ARCHIVE-FILTER-XZ       
   ARCHIVE-FILTER-UU       
   ARCHIVE-FILTER-RPM      
   ARCHIVE-FILTER-LZIP     
   ARCHIVE-FILTER-LRZIP   
   ARCHIVE-FILTER-LZOP    
   ARCHIVE-FILTER-GRZIP   
   ARCHIVE-FILTER-LZ4     

   ARCHIVE-FORMAT-BASE-MASK        
   ARCHIVE-FORMAT-CPIO                
   ARCHIVE-FORMAT-CPIO-POSIX          
   ARCHIVE-FORMAT-CPIO-BIN-LE         
   ARCHIVE-FORMAT-CPIO-BIN-BE         
   ARCHIVE-FORMAT-CPIO-SVR4-NOCRC     
   ARCHIVE-FORMAT-CPIO-SVR4-CRC       
   ARCHIVE-FORMAT-CPIO-AFIO-LARGE     
   ARCHIVE-FORMAT-SHAR               
   ARCHIVE-FORMAT-SHAR-BASE          
   ARCHIVE-FORMAT-SHAR-DUMP          
   ARCHIVE-FORMAT-TAR                
   ARCHIVE-FORMAT-TAR-USTAR          
   ARCHIVE-FORMAT-TAR-PAX-INTERCHANGE
   ARCHIVE-FORMAT-TAR-PAX-RESTRICTED 
   ARCHIVE-FORMAT-TAR-GNUTAR         
   ARCHIVE-FORMAT-ISO9660            
   ARCHIVE-FORMAT-ISO9660-ROCKRIDGE  
   ARCHIVE-FORMAT-ZIP                
   ARCHIVE-FORMAT-EMPTY              
   ARCHIVE-FORMAT-AR                 
   ARCHIVE-FORMAT-AR-GNU             
   ARCHIVE-FORMAT-AR-BSD             
   ARCHIVE-FORMAT-MTREE              
   ARCHIVE-FORMAT-RAW                
   ARCHIVE-FORMAT-XAR                
   ARCHIVE-FORMAT-LHA                
   ARCHIVE-FORMAT-CAB                
   ARCHIVE-FORMAT-RAR                
   ARCHIVE-FORMAT-7ZIP               
   ARCHIVE-FORMAT-WARC               


   ))



(select-module util.archives)
(dynamic-load "archive")

(define-condition-type <archive-error> <io-error>
  archive-error?
  (debug-info archive-error-debug-info)
  (message archive-error-message))

(define-condition-type <archive-reader-error> <archive-error> 
  archive-reader-error?
  (debug-info archive-reader-error-debug-info)
  (message archive-reader-error-message))

(define-condition-type <archive-writer-error> <archive-error> 
  archive-writer-error?
  (debug-info archive-writer-error-debug-info)
  (message archive-writer-error-message))

(define-condition-type <archive-builder-error> <archive-error> 
  archive-builder-error?
  (debug-info archive-builder-error-debug-info)
  (message archive-builder-error-message))

(define-condition-type <archive-entry-error> <archive-error> 
  archive-entry-error?
  (debug-info archive-entry-error-debug-info)
  (message archive-entry-error-message))

(define-syntax archive-throw
  (syntax-rules ()
    ((_ (co r d))
     (raise
      (condition
       (co (message r) (debug-info d)))))
    ((_ ((co r d) ...))
     (raise
      (condition
       (co (message r) (debug-info d)) ...)))))

(define-syntax archive-try
  (syntax-rules ()
    ((_ condition-object body ...)
     (receive (code msg dbg data)
         body ...
         (if (not (= code 0))
           (archive-throw (condition-object msg dbg))
           data)))
    ((_ . ?) (syntax-error "malformed syntax "
                           (archive-try condition-object body)))))



(define-class <archive-reader> (<instance-pool-mixin> <collection>)
  ((type :init-form 'file)    ;;; 'file or 'memory
   (name :init-form "")       ;;; archive name
   (data :init-form #f)  
   (archive-core :init-form (make <archive-base>))
   (status :allocation :virtual
           :slot-ref (lambda (o)
                       (receive (code str debug data)
                           (%is-read-archive-open o)
                         (if (= code 0) 'close 'open))))))

(define-class <archive-writer> ()
  ((status :init-form 'close) ;;; 'open or 'close
   (name :init-form "")       ;;; archive name
   (format-type :init-form ARCHIVE-FORMAT-TAR-GNUTAR)
   (filter-type :init-form ARCHIVE-FILTER-NONE)
   (append-list :init-form '())
   (remove-list :init-form '())
   (modify-list :init-form '())))

(define-class <archive-builder> ()
  ((type :init-form 'file) ;; 'file or 'memory
   (name :init-form "")    ;; If type is 'memory always empty.
   (format-type :init-form ARCHIVE-FORMAT-TAR-GNUTAR)
   (filter-type :init-form ARCHIVE-FILTER-NONE)
   (entry-list :init-form '())))

(define-class <archive-entry> (<virtual-output-port>)
  ((isdir :init-form #f)
   (uid :init-form 1000)
   (gid :init-form 1000)
   (mode :init-form 0)
   (atime :init-form 0)
   (ctime :init-form 0)
   (mtime :init-form 0)
   (name :init-form "")
   (size :init-form 0)
   (type :init-form 'file) ;; 'file, 'arcchive or 'memory
   (data :init-form #f)
   (data-opener :init-form (^() #f))
   (data-closer :init-form (^() #f))
   (data-getter :init-form (^() #f))
   (data-flusher :init-form (^(new-out :optional opener closer getter) #f))))



(define (open-input-archive-from-name name)
  (let ((obj (make <archive-reader>)))
    (set! (~ obj 'name) name)
    obj))

(define (open-input-archive-from-memory oport)
  (let ((obj (make <archive-reader>)))
    (set! (~ obj 'type) 'memory)
    (set! (~ obj 'data) oport)
    obj))

(define-method close-input-archive ((cls <archive-reader>))
  (instance-pool-remove! <archive-reader>
                         (^x (eq? x cls)))
  (archive-try <archive-reader-error>
               (%free-read-archive (~ cls 'archive-core))))

(define (make-archive-entry :key (buffer #f) (opener open-output-string) (closer close-output-port) (getter get-output-string) (putb-func write-byte))
  (define (func-setter obj)
    (set! (~ obj 'data-opener) opener)
    (set! (~ obj 'data-closer) closer)
    (set! (~ obj 'data-getter) getter))
  (let* ((out (or buffer (opener)))
         (obj (make <archive-entry>
                :putb (lambda (b) (putb-func b out)))))
    (set! (~ obj 'data) out)
    (func-setter obj)
    (set! (~ obj 'data-flusher)
          (^(new-out :optional (opener open-output-string)
                     (closer close-output-port) (getter get-output-string))
            ((~ obj 'data-closer) out)
            (set! out (or new-out (opener)))
            (set! (~ obj 'data) out)
            (func-setter obj)))
    obj))

(define-method get-archive-entry-data ((cls <archive-entry>))
  (if #0=(~ cls 'data) ((~ cls 'data-getter) #0#) ""))

(define-method flush-archive-entry-data ((cls <archive-entry>) new-out :optional (opener open-output-string) (closer close-output-port) (getter get-output-string))
  (begin0
      (get-archive-entry-data cls)
    ((~ cls 'data-flusher) new-out opener closer getter)))

(define (create-entry-from-file name :optional (opener open-output-string) (closer close-output-port) (getter get-output-string))
  (let ((entry #f)
        (stat #f))
    (if (eq? (file-exists? name) #f)
      (archive-throw
       (<archive-entry-error> #"Can't find ~name" "create-entry-from-file")))
    (set! entry (make-archive-entry :opener opener :closer closer :getter getter))
    (set! stat (sys-stat name))
    (set! (~ entry 'type) 'file)
    (set! (~ entry 'uid) (sys-getuid))
    (set! (~ entry 'gid) (sys-getgid))
    (set! (~ entry 'mode) (~ stat 'mode))
    (set! (~ entry 'atime) (~ stat 'atime))
    (set! (~ entry 'ctime) (~ stat 'ctime))
    (set! (~ entry 'mtime) (~ stat 'mtime))
    (set! (~ entry 'name) name)
    (set! (~ entry 'size) (file-size name))
    entry))

(define (create-entry-from-memory name :optional (buffer #f) (opener open-output-string) (closer close-output-port) (getter get-output-string))
  (let ((entry #f))
    (set! entry (make-archive-entry :buffer buffer :opener opener :closer closer :getter getter))
    (set! (~ entry 'type) 'memory)
    (set! (~ entry 'uid) (sys-getuid))
    (set! (~ entry 'gid) (sys-getgid))
    (set! (~ entry 'mode) 0)
    (set! (~ entry 'atime) (sys-time))
    (set! (~ entry 'ctime) (sys-time))
    (set! (~ entry 'mtime) (sys-time))
    (set! (~ entry 'name) name)
    (set! (~ entry 'size) (if buffer (port-seek buffer 0 SEEK_CUR) 0))
    entry))

(define-method get-entry ((cls <archive-reader>))
  (let ((entry (make-archive-entry)))
    (set! (~ entry 'type) 'archive)
    (archive-try <archive-reader-error>
                 (%get-entry cls entry))
    entry))

(define-method read-file ((cls <archive-entry>))
  (let ((iport #f))
    (if (or (not (eq? (~ cls 'type) 'file))
            (string=? (~ cls 'name) ""))
      (archive-throw
       (<archive-entry-error> "Inappropriate <archive-entry>" "read-file")))
    (begin
      (set! iport (open-input-file (~ cls 'name)))
      (copy-port iport (~ cls 'data))
      (close-port iport))))

(define-method read-data ((cls <archive-reader>) (entry <archive-entry>))
  (if (and (eq? (~ cls 'status) 'close)
           (eq? (~ entry 'type) 'archive))
    (archive-throw
     (<archive-entry-error>
      "This function is available while archive iteration only, thus archive already must be opened."
      "read-data")))
  (archive-try <archive-entry-error>
               (%read-data cls entry)))

(define (re-calc-file-size lst)
  (dolist (entry lst)
    (cond
     ((eq? (~ entry 'type) 'file)
      (set! (~ entry 'size) (file-size (~ entry 'name))))
     ((or (eq? (~ entry 'type) 'memory)
          (eq? (~ entry 'type) 'archive))
      (set! (~ entry 'size) (port-seek (~ entry 'data) 0 SEEK_CUR))))))

(define-method extract-all-files ((cls <archive-reader>))
  (archive-try <archive-reader-error>
               (%extract-all-file cls)))

(define-method extract-files-to-disk ((cls <archive-reader>) entry-list)
  (archive-try <archive-reader-error>
               (%extract-file-to-disk cls)))

(define-method extract-files-to-memory ((cls <archive-reader>) entry-list)
  (archive-try <archive-reader-error>
               (%extract-file-to-memory cls)))

;;; auxiliary method
(define-method begin-iterator ((cls <archive-reader>))
  (cond
   ((eq? (~ cls 'type) 'file)
    (archive-try <archive-reader-error>
                 (%open-read-archive-from-filename cls (~ cls 'name))))
   ((eq? (~ cls 'type) 'memory)
    (archive-try <archive-reader-error>
                 (%open-read-archive-from-memory cls (~ cls 'data))))))

(define-method end-iterator ((cls <archive-reader>))
  (archive-try <archive-reader-error>
               (%close-read-archive cls)))

(define-method call-with-iterator ((cls <archive-reader>) proc . opts)
  (let ((entry #f)
        (ret #t))
    (begin-iterator cls)
    (proc
     (lambda ()
       (set! entry (make-archive-entry))
       (set! (~ entry 'type) 'archive)
       (receive (code msg dbg data)
           (%get-entry cls entry)
         (set! ret (list code msg dbg data)))
       (if (eq? ret #t) #f
           (begin
             (cond
              ((eq? (car ret) #f)
               (end-iterator cls)
               #t)
              ((= (car ret) 0) #f)
              (t
               (match ret
                 ((code msg dbg data)
                  (begin
                    (end-iterator cls)
                    (archive-try <archive-reader-error>
                                 (values code msg dbg data))))))))))
     (lambda () entry))))

(define (call-with-file-archive name proc)
  (let ((archive (open-input-archive-from-name name)))
    (unwind-protect
     (proc archive)
     (close-input-archive archive))))

(define (call-with-memory-archive oport proc)
  (let ((archive (open-input-archive-from-memory oport)))
    (unwind-protect
     (proc archive)
     (close-input-archive archive))))



(define (create-archive archive-name entry-list format-type :optional (filter-type ARCHIVE-FILTER-NONE))
  (archive-try
   <archive-writer-error>
   (%create-archive archive-name entry-list format-type filter-type)))

(define (open-output-archive name format-type :key (filter-type ARCHIVE-FILTER-NONE))
  (let ((writer #f))
    (if (or (eq? (file-exists? name) #f)
            (= (file-size name) 0))
      (archive-throw
       (<archive-writer-error>
        "Archive is not exists or size = 0" "open-output-archive")))
    (set! writer (make <archive-writer>))
    (set! (~ writer 'status) 'open)
    (set! (~ writer 'name) name)
    (set! (~ writer 'format-type) format-type)
    (set! (~ writer 'filter-type) filter-type)
    writer))



(define-method close-output-archive ((cls <archive-writer>))
  (define (if-exists-in-list org dest)
    (not (every (^x (eq? x #f))
                (let ((ret '()))
                  (dolist (v dest (reverse ret))
                    (push! ret
                           (find (^(x)
                                   (and (not (eq? x #f))
                                        (string=? (~ x 'name) (~ v 'name))))
                                 org)))))))
  (define (normalize-list lst)
    (delete-duplicates! lst 
                        (lambda (x y)
                          (string=? (~ x 'name) (~ y 'name))))
    (filter (^x (not (eq? x #f))) lst))
  (define (is-archive-already-allocate? name)
    (find (^x (string=? (~ x 'name) name))
          (instance-pool->list <archive-reader>)))
  (let ((original-list '()) 
        (tmp-name "")
        (ar #f))
    (if (not (eq? (is-archive-already-allocate? (~ cls 'name)) #f))
      (archive-throw
       (<archive-writer-error>
        #"~(~ cls 'name) was already opened. Please close it before writing it."
        "close-output-archive")))
    (set! ar (open-input-archive-from-name (~ cls 'name)))
    (set! original-list (map identity ar))
    (close-input-archive ar)
    (set! #0=(~ cls 'append-list) (normalize-list #0#))
    (set! #1=(~ cls 'remove-list) (normalize-list
                                   (append #1# (~ cls 'modify-list))))
    (set! #2=(~ cls 'modify-list) (normalize-list #2#))
    (if (if-exists-in-list original-list #0#)
      (archive-throw
       ((<archive-entry-error>
         "file for append is already exists in archive."
         "close-output-archive")
        (<archive-writer-error>
         "Can't continue writing archive." "close-output-archive"))))
    (re-calc-file-size (~ cls 'append-list))
    (re-calc-file-size (~ cls 'remove-list))
    (re-calc-file-size (~ cls 'modify-list))

    (set! tmp-name (sys-basename (sys-tmpnam)))
    (%copy-write-archive tmp-name (~ cls 'name)
                         (~ cls 'remove-list)
                         (append (~ cls 'append-list) (~ cls 'modify-list))
                         (~ cls 'format-type) (~ cls 'filter-type))
    (move-file (~ cls 'name)
               (format #f "~a.~a"
                       (~ cls 'name)
                       ($ sys-strftime "%Y%m%d%H%M%S" $ sys-localtime $ sys-time)))
    (move-file tmp-name (~ cls 'name))))


(define-method append-entry ((cls <archive-writer>) entry)
  (set! #0=(~ cls 'append-list) (append (list entry) #0#)))

(define-method remove-entry ((cls <archive-writer>) entry)
  (set! #0=(~ cls 'remove-list) (append (list entry) #0#)))

(define-method modify-entry ((cls <archive-writer>) entry)
  (set! #0=(~ cls 'modify-list) (append (list entry) #0#)))

(define (open-build-archive-to-file name format-type :optional (filter-type ARCHIVE-FILTER-NONE))
  (let ((obj (make <archive-builder>)))
    (set! (~ obj 'name) name)
    (set! (~ obj 'format-type) format-type)
    (set! (~ obj 'filter-type) filter-type)
    obj))

(define (open-build-archive-to-memory format-type :optional (filter-type ARCHIVE-FILTER-NONE))
  (let ((obj (make <archive-builder>)))
    (set! (~ obj 'type) 'memory)
    (set! (~ obj 'format-type) format-type)
    (set! (~ obj 'filter-type) filter-type)
    obj))

(define-method close-build-archive ((cls <archive-builder>))
  (let ((options (option->string cls)))
    (cond
     ((eq? (~ cls 'type) 'file) (close-file-build-archive cls options))
     ((eq? (~ cls 'type) 'memory) (close-memory-build-archive cls options))
     (t (archive-throw
         (<archive-builder-error> #"Unknown archive type of ~(~ cls 'name)."
                                  "close-build-archive"))))))

(define (close-file-build-archive archive option-string)
  (re-calc-file-size (~ archive 'entry-list))
  (archive-try <archive-builder-error>
               (%write-archive-to-disk
                (~ archive 'name) (~ archive 'entry-list)
                (~ archive 'format-type) (~ archive 'filter-type)
                option-string)))

(define (close-memory-build-archive archive option-string)
  (let ((size 0))
    (re-calc-file-size (~ archive 'entry-list))
    (set! size (fold (^(x y) (+ (~ x 'size) y))
                     size (~ archive 'entry-list)))
    (archive-try
     <archive-builder-error>
     (%write-archive-to-memory
      size (~ archive 'entry-list)
      (~ archive 'format-type) (~ archive 'filter-type)
      option-string))))

(define-method append-entry ((cls <archive-builder>) entry)
  (set! #0=(~ cls 'entry-list) (reverse (append (list entry) #0#))))



;;;;;;;;;;;;;;;
;;; validators
;;;;;;;;;;;;;;;
(define (compression-level-validator obj value)
  (if (and (integer? value) (<= 0 value 9))
    value
    (error "integer number is required or one of value from 0 to 9, at `compression-level` slot.")))

(define (boolean-validator obj value)
  (if (boolean? value)
    (booleans->integer value)
    (error "boolean value is required.")))

(define (7zip-compress-validator obj value)
  (define compress-names '("bzip2" "copy" "deflate" "lzma1" "lzma2" "ppmd"))
  (if (and (string? value)
           (member value compress-names))
    value
    (error (format #f "string is required or one of value in ~a, at 7zip:compress slot." (string-join (map (^x (format #f "'~a'" x)) compress-names) ", ")))))

(define (string-validator obj value)
  (if (string? value)
    value
    (error "string is required.")))

(define (xar-validator obj value)
  (if (string? value)
    (cond
     ((string=? value "") "0")
     ((member value '("sha1" "md5")) value))
    (error "string is required.")))

(define (xar-compress-validator obj value)
  (define compress-names '("gzip" "bzip2" "lzma" "xz"))
  (if (and (string? value)
           (member value compress-names))
    value
    (error (format #f "string is required or one of value in ~a"
                   (string-join (map (^x (format #f "'~a'" x))
                                     compress-names)
                                ", ")))))

(define (zip-encrypt-validator obj value)
  (if (and (string=? value)
           (member '("aes128" "aes256")))
    value
    (error "string is require or one of value in 'aes128', 'aes256'.")))

(define (zip-compression-validator obj value)
  (define compress-names '("deflate" "store"))
  (if (and (string? value)
           (member value compress-names))
    value
    (error (format #f "string is required or one of value in ~a, at 7zip:compress slot." (string-join (map (^x (format #f "'~a'" x)) compress-names) ", ")))))

(define-method option->string ((cls <archive-builder>))
  (let* ((base-slots (map slot-definition-name (class-slots <archive-builder>)))
         (deriv-slots (map slot-definition-name (class-slots (class-of cls))))
         (option-slots (filter (^x (not (member x base-slots))) deriv-slots)))
    (map (^x (string-join x "="))
         (map (^x (list (x->string (car x)) (x->string (cadr x))))
         (zip option-slots (map (^x (slot-ref cls x)) option-slots))))))

(define *archive-filter-alist* '())

(define (generate-symbol fmt name)
  (string->symbol (format #f fmt name)))

(define (generate-super-class-list type pre-class)
  (if #0=(get-archive-filter-class type)
      (list (class-name <archive-builder>) pre-class #0#)
      (list (class-name <archive-builder>) pre-class)))

(define (get-archive-filter-class type)
  (if #0=(assv type *archive-filter-alist*)
      (class-name (cdr #0#))
      #f))

(define (generate-slot-definition slot-definitions)
  (map (^x
         (match x
           ((name init-value validator)
            (list name :init-keyword (make-keyword name)
                  :init-form init-value
                  :validator validator))))
       slot-definitions))

(define-macro (define-archive-option-mixin type name slot-definitions)
  `(begin
     (define-class #0=,(string->symbol (format #f "<~a>" name)) (<validator-mixin>)
       ,(generate-slot-definition slot-definitions))
     (set! *archive-filter-alist*
           (append (list (cons ,type #0#)) *archive-filter-alist*))))

(define-macro (define-archive-builder-class base-name format-type :optional (filter-type ARCHIVE-FILTER-NONE) (mixin-type ARCHIVE-FILTER-NONE))
  `(begin
     (define-class #0=,(generate-symbol "<archive-~a-builder-meta>" base-name)
       (<class>) ())
     (define-class #1=,(generate-symbol "<archive-~a-builder-pre>" base-name)
       () ()
       :metaclass #0#)
     (define-class #2=,(generate-symbol "<archive-~a-builder>" base-name)
       ,(generate-super-class-list (eval mixin-type ())
                                   (generate-symbol
                                    "<archive-~a-builder-pre>" base-name))
       ())

     (define (,(generate-symbol "open-build-~a-archive-to-file" base-name)
              name)
       (let ((obj (make #2#)))
         (set! (~ obj 'name) name)
         (set! (~ obj 'format-type) ,format-type)
         (set! (~ obj 'filter-type) ,filter-type)
         obj))

     (define (,(generate-symbol "open-build-~a-archive-to-memory" base-name))
       (let ((obj (make #2#)))
         (set! (~ obj 'type) 'memory)
         (set! (~ obj 'format-type) ,format-type)
         (set! (~ obj 'filter-type) ,filter-type)
         obj))

     (define-method call-with-builder ((cls #0#) proc . opts)
       (let ((obj (make #2#)))
         (set! (~ obj 'name) "")
         (set! (~ obj 'type) 'memory)
         (set! (~ obj 'format-type) ,format-type)
         (set! (~ obj 'filter-type) ,filter-type)
         (proc
          (lambda (src)
            (if (string? src)
              (let ((entry (create-entry-from-file src)))
                (append-entry obj entry))))
          (lambda () obj))))
     (export #0# #1# #2# call-with-builder)))


(define-archive-option-mixin ARCHIVE-FILTER-GZIP "filter-gzip-mixin"
  ((gzip:compression-level 9 compression-level-validator)
   (gzip:timestamp #t boolean-validator)))

(define-archive-option-mixin ARCHIVE-FILTER-BZIP2 "filter-bzip2-mixin"
  ((bzip2:compression-level 9 compression-level-validator)))

(define-archive-option-mixin ARCHIVE-FILTER-LZ4 "filter-lz4-mixin"
  ((lz4:compression-level 9 compression-level-validator)))

(define-archive-option-mixin ARCHIVE-FILTER-LZIP "filter-lzip-mixin"
  ((lzip:compression-level 9 compression-level-validator)))

(define-archive-option-mixin ARCHIVE-FILTER-LZMA "filter-lzma-mixin"
  ((lzma:compression-level 9 compression-level-validator)))

(define-archive-option-mixin ARCHIVE-FILTER-LZOP "filter-lzop-mixin"
  ((lzop:compression-level 9 compression-level-validator)))

(define-archive-option-mixin ARCHIVE-FILTER-UU "filter-uuencode-mixin"
  ((uuencode:name "" (^(o v) v))
   (uuencode:mode "0644" (^(o v) v))))

(define-archive-option-mixin ARCHIVE-FILTER-XZ "filter-xz-mixin"
  ((xz:compression-level 9 compression-level-validator)))

(define-archive-option-mixin ARCHIVE-FORMAT-7ZIP "format-7zip-mixin"
  ((7zip:compression "copy" 7zip-compress-validator)))

(define-archive-option-mixin ARCHIVE-FORMAT-ISO9660 "format-is09660-mixin"
  ((iso9660:boot "boot.img" string-validator)
   (iso9660:boot-info-table #t boolean-validator)))

(define-archive-option-mixin ARCHIVE-FORMAT-MTREE "format-mtree-mixin"
  ((mtree:use-set #t boolean-validator)
   (mtree:dironly #t boolean-validator)
   (mtree:all #t boolean-validator)
   (mtree:flags #t boolean-validator)
   (mtree:type #t boolean-validator)))

(define-archive-option-mixin ARCHIVE-FORMAT-XAR "format-xar-mixin"
  ((xar:toc-checksum "" xar-validator)
   (xar:checksum "" xar-validator)
   (xar:compression "gzip" xar-compress-validator)
   (xar:compression-level 9 compression-level-validator)))

(define-archive-option-mixin ARCHIVE-FORMAT-ZIP "format-zip-mixin"
  ((zip:experimentral #t boolean-validator)
   (zip:zip64 #f boolean-validator)
   ;; (zip:encryption "" zip-encrypt-validator)
   (zip:compression "deflate" string-validator)
   (zip:compression-level 9 compression-level-validator)
   (zip:ignorecrc32 #f boolean-validator)
   (zip:fakecrc32 #f boolean-validator)))

(define-method archive-file-builder->memory-builder ((cls <archive-builder>))
  (set! (~ cls 'type) 'memory)
  (set! (~ cls 'name) ""))

(define-method archive-memory-builder->file-builder ((cls <archive-builder>) name)
  (set! (~ cls 'type) 'file)
  (set! (~ cls 'name) name))


(define-archive-builder-class "gnu-tar-gz" ARCHIVE-FORMAT-TAR-GNUTAR ARCHIVE-FILTER-GZIP ARCHIVE-FILTER-GZIP)
(define-archive-builder-class "gnu-tar-bz2" ARCHIVE-FORMAT-TAR-GNUTAR ARCHIVE-FILTER-BZIP2 ARCHIVE-FILTER-BZIP2)
(define-archive-builder-class "gnu-tar-lzma" ARCHIVE-FORMAT-TAR-GNUTAR ARCHIVE-FILTER-LZMA ARCHIVE-FILTER-LZMA)
(define-archive-builder-class "gnu-tar-uu" ARCHIVE-FORMAT-TAR-GNUTAR ARCHIVE-FILTER-UU ARCHIVE-FILTER-UU)
(define-archive-builder-class "gnu-tar-xz" ARCHIVE-FORMAT-TAR-GNUTAR ARCHIVE-FILTER-XZ ARCHIVE-FILTER-XZ)
(define-archive-builder-class "us-tar-gz" ARCHIVE-FORMAT-TAR-USTAR ARCHIVE-FILTER-GZIP ARCHIVE-FILTER-GZIP)
(define-archive-builder-class "us-tar-bz" ARCHIVE-FORMAT-TAR-USTAR ARCHIVE-FILTER-BZIP2 ARCHIVE-FILTER-BZIP2)
(define-archive-builder-class "us-tar-lzma" ARCHIVE-FORMAT-TAR-USTAR ARCHIVE-FILTER-LZMA ARCHIVE-FILTER-LZMA)
(define-archive-builder-class "us-tar-uu" ARCHIVE-FORMAT-TAR-USTAR ARCHIVE-FILTER-UU ARCHIVE-FILTER-UU)
(define-archive-builder-class "us-tar-xz" ARCHIVE-FORMAT-TAR-USTAR ARCHIVE-FILTER-XZ ARCHIVE-FILTER-XZ)
(define-archive-builder-class "pax-tar-gz" ARCHIVE-FORMAT-TAR-PAX-RESTRICTED ARCHIVE-FILTER-GZIP ARCHIVE-FILTER-GZIP)
(define-archive-builder-class "pax-tar-bz" ARCHIVE-FORMAT-TAR-PAX-RESTRICTED ARCHIVE-FILTER-BZIP2 ARCHIVE-FILTER-BZIP2)
(define-archive-builder-class "pax-tar-lzma" ARCHIVE-FORMAT-TAR-PAX-RESTRICTED ARCHIVE-FILTER-LZMA ARCHIVE-FILTER-LZMA)
(define-archive-builder-class "pax-tar-uu" ARCHIVE-FORMAT-TAR-PAX-RESTRICTED ARCHIVE-FILTER-UU ARCHIVE-FILTER-UU)
(define-archive-builder-class "pax-tar-xz" ARCHIVE-FORMAT-TAR-PAX-RESTRICTED ARCHIVE-FILTER-XZ ARCHIVE-FILTER-XZ)
(define-archive-builder-class "zip" ARCHIVE-FORMAT-ZIP ARCHIVE-FILTER-NONE ARCHIVE-FORMAT-ZIP)
(define-archive-builder-class "7zip" ARCHIVE-FORMAT-7ZIP ARCHIVE-FILTER-NONE ARCHIVE-FORMAT-7ZIP)

;; (guard
;;     (exc
;;      ((archive-error? exc)
;;       (format #t "~a ~a\n"
;;               (archive-error-debug-info exc)
;;               (archive-error-reason exc))))
;;   (let ((obj (coerce-to <archive-gnu-tar-gz-builder> '("options.scm"))))
;;     (archive-memory-builder->file-builder obj "foo2.tar.gz")
;;     (set! (~ obj 'gzip:compression-level) 0)
;;     (describe obj)
;;     (close-build-archive obj)))
