
#include "archive_proc.h"

ScmObj open_read_archive( ScmObj archive_name )
{
     Scm_libarchive_archive *a = NULL;
     int                     r = ARCHIVE_OK;

     a = SCM_LIBARCHIVE_ARCHIVE( libarchive_archive_allocate( NULL, NULL ) );
     a->obj = archive_read_new();
     archive_read_support_filter_all( a->obj );
     archive_read_support_format_all( a->obj );
     r = archive_read_open_filename(
         a->obj, Scm_GetString( SCM_STRING( archive_name ) ), BUFSIZ );
     if( r != ARCHIVE_OK ) {
          Scm_Error( archive_error_string( a->obj ) );
     }

     a->name = archive_name;
     a->status = Scm_MakeSymbol( SCM_STRING( SCM_MAKE_STR( "open" ) ), 1 );
     a->type = GAUCHE_ARCHIVE_READ;
     return SCM_OBJ( a );
}

ScmObj close_read_archive( ScmObj archive )
{
     Scm_libarchive_archive *a = NULL;

     a = SCM_LIBARCHIVE_ARCHIVE( archive );
     if( a->obj ) {
          archive_read_close( a->obj );
          archive_read_free( a->obj );
     }
     a->name = SCM_MAKE_STR( "" );
     a->status = Scm_MakeSymbol( SCM_STRING( SCM_MAKE_STR( "close" ) ), 1 );
     a->type = GAUCHE_ARCHIVE_NONE;

     return SCM_TRUE;
}

ScmObj get_entry( ScmObj archive )
{
     Scm_libarchive_archive *      a = NULL;
     Scm_libarchive_archive_entry *entry = NULL;
     struct archive_entry *        _entry = NULL;
     int                           r = ARCHIVE_OK;

#ifdef HAVE_SYS_STAT_H
     const struct stat *st = NULL;
#endif /* #ifdef HAVE_SYS_STAT_H */

     a = SCM_LIBARCHIVE_ARCHIVE( archive );
     entry = SCM_LIBARCHIVE_ARCHIVE_ENTRY(
         libarchive_archive_entry_allocate( NULL, NULL ) );

     r = archive_read_next_header( a->obj, &_entry );
     if( r == ARCHIVE_OK ) {
          entry->archive_name =
              strdup( Scm_GetString( SCM_STRING( a->name ) ) );
          entry->entry_name = strdup( archive_entry_pathname_utf8( _entry ) );
          /* printf( "[[%s]]\n", entry->entry_name ); */
#ifdef HAVE_SYS_STAT_H
          st = archive_entry_stat( _entry );

          if( S_ISREG( st->st_mode ) ) {
               entry->isdir = SCM_FALSE;
          } else if( S_ISDIR( st->st_mode ) ) {
               entry->isdir = SCM_TRUE;
          }
          entry->uid = SCM_MAKE_INT( st->st_uid );
          entry->gid = SCM_MAKE_INT( st->st_gid );
          entry->mode = SCM_MAKE_INT( st->st_mode );
          entry->atime = SCM_MAKE_INT( st->st_atim.tv_sec );
          entry->ctime = SCM_MAKE_INT( st->st_ctim.tv_sec );
          entry->mtime = SCM_MAKE_INT( st->st_mtim.tv_sec );
          entry->size = SCM_MAKE_INT( st->st_size );
#endif /* #ifdef HAVE_SYS_STAT_H */
          entry->name =
              SCM_MAKE_STR_COPYING( archive_entry_pathname_utf8( _entry ) );

          return SCM_OBJ( entry );
     }

     return SCM_FALSE;
}

int find_name( struct archive_entry *entry, ScmObj lst )
{
     ScmObj                        p;
     const char *                  src = NULL;
     const char *                  dest = NULL;
     Scm_libarchive_archive_entry *_entry = NULL;

     src = archive_entry_pathname_utf8( entry );
     SCM_FOR_EACH( p, lst )
     {
          _entry = SCM_LIBARCHIVE_ARCHIVE_ENTRY( SCM_CAR( p ) );
          dest = _entry->entry_name;
          /* fprintf( stderr, "src:[%s] dest:[%s]\n", src, dest ); */
          if( strcmp( src, dest ) == 0 ) {
               return 1;
          }
     }

     return 0;
}

ScmObj extract_file_to_disk( ScmObj entry_list )
{
     int                           flags = 0;
     int                           r = 0;
     long                          len = 0;
     char                          buf[BUFSIZ] = { '\0' };
     Scm_libarchive_archive_entry *p = NULL;
     struct archive *              ext = NULL;
     struct archive *              a = NULL;
     struct archive_entry *        entry_iter = NULL;

     p = SCM_LIBARCHIVE_ARCHIVE_ENTRY( SCM_CAR( entry_list ) );
     /*
      * fprintf( stderr, "[%s]\n", p->archive_name );
      * fprintf( stderr, "[%s]\n", p->entry_name );
      */

     flags = ARCHIVE_EXTRACT_TIME;
     flags |= ARCHIVE_EXTRACT_PERM;
     flags |= ARCHIVE_EXTRACT_ACL;
     flags |= ARCHIVE_EXTRACT_FFLAGS;

     a = archive_read_new();
     archive_read_support_filter_all( a );
     archive_read_support_format_all( a );
     ext = archive_write_disk_new();
     archive_write_disk_set_options( ext, flags );
     archive_write_disk_set_standard_lookup( ext );

     if( ( r = archive_read_open_filename( a, p->archive_name, BUFSIZ ) ) ) {
          Scm_Error( archive_error_string( a ) );
     }

     for( ;; ) {
          r = archive_read_next_header( a, &entry_iter );
          if( r == ARCHIVE_EOF ) {
               break;
          }
          if( r < ARCHIVE_OK ) {
               Scm_Error( archive_error_string( a ) );
          }
          if( r < ARCHIVE_WARN ) {
               Scm_Error( archive_error_string( a ) );
          }

          if( find_name( entry_iter, entry_list ) == 1 ) {
               r = archive_write_header( ext, entry_iter );
               if( r < ARCHIVE_OK ) {
                    Scm_Error( archive_error_string( ext ) );
               }
               if( archive_entry_size( entry_iter ) > 0 ) {
                    len = archive_read_data( a, buf, sizeof( buf ) );
                    while( len > 0 ) {
                         if( archive_write_data( ext, buf, len ) != len )
                              Scm_Error( archive_error_string( ext ) );
                         len = archive_read_data( a, buf, sizeof( buf ) );
                    }
                    if( len < 0 ) {
                         Scm_Error( archive_error_string( a ) );
                    }
               }

               r = archive_write_finish_entry( ext );
               if( r < ARCHIVE_OK ) {
                    Scm_Error( archive_error_string( ext ) );
               }
               if( r < ARCHIVE_WARN ) {
                    Scm_Error( archive_error_string( ext ) );
               }
          }
     }

     archive_read_close( a );
     archive_read_free( a );
     archive_write_close( ext );
     archive_write_free( ext );

     return SCM_TRUE;
}

ScmObj extract_file_to_memory( ScmObj entry_list )
{
     int                           flags = 0;
     int                           r = 0;
     long                          len = 0;
     char                          buf[BUFSIZ] = { '\0' };
     Scm_libarchive_archive_entry *p = NULL;
     struct archive *              ext = NULL;
     struct archive *              a = NULL;
     struct archive_entry *        entry_iter = NULL;
     ScmObj                        result_list = SCM_NIL;
     ScmObj                        sport;

     p = SCM_LIBARCHIVE_ARCHIVE_ENTRY( SCM_CAR( entry_list ) );
     /*
      * fprintf( stderr, "[%s]\n", p->archive_name );
      * fprintf( stderr, "[%s]\n", p->entry_name );
      */

     flags = ARCHIVE_EXTRACT_TIME;
     flags |= ARCHIVE_EXTRACT_PERM;
     flags |= ARCHIVE_EXTRACT_ACL;
     flags |= ARCHIVE_EXTRACT_FFLAGS;

     a = archive_read_new();
     archive_read_support_filter_all( a );
     archive_read_support_format_all( a );
     ext = archive_write_disk_new();
     archive_write_disk_set_options( ext, flags );
     archive_write_disk_set_standard_lookup( ext );

     if( ( r = archive_read_open_filename( a, p->archive_name, BUFSIZ ) ) ) {
          Scm_Error( archive_error_string( a ) );
     }

     for( ;; ) {
          r = archive_read_next_header( a, &entry_iter );
          if( r == ARCHIVE_EOF ) {
               break;
          }
          if( r < ARCHIVE_OK ) {
               Scm_Error( archive_error_string( a ) );
          }
          if( r < ARCHIVE_WARN ) {
               Scm_Error( archive_error_string( a ) );
          }

          if( find_name( entry_iter, entry_list ) == 1 ) {
               /*
                * r = archive_write_header( ext, entry_iter );
                * if (r < ARCHIVE_OK) {
                *      Scm_Error(archive_error_string(ext));
                * }
                */
               if( archive_entry_size( entry_iter ) > 0 ) {
                    sport = Scm_MakeOutputStringPort( 1 );
                    len = archive_read_data( a, buf, sizeof( buf ) );
                    while( len > 0 ) {
                         /*
                          * if (archive_write_data(ext, buf, len) != len)
                          *      Scm_Error(archive_error_string(ext));
                          */
                         Scm_Putz( buf, len, SCM_PORT( sport ) );
                         len = archive_read_data( a, buf, sizeof( buf ) );
                    }
                    if( len < 0 ) {
                         Scm_Error( archive_error_string( a ) );
                    }
                    result_list =
                        Scm_Append2( result_list, SCM_LIST1( sport ) );
               }

               r = archive_write_finish_entry( ext );
               if( r < ARCHIVE_OK ) {
                    Scm_Error( archive_error_string( ext ) );
               }
               if( r < ARCHIVE_WARN ) {
                    Scm_Error( archive_error_string( ext ) );
               }
          }
     }

     archive_read_close( a );
     archive_read_free( a );
     archive_write_close( ext );
     archive_write_free( ext );

     return result_list;
}

ScmObj extract_all_file( ScmObj archive )
{
     int                     flags = 0;
     int                     r = 0;
     long                    len = 0;
     char                    buf[BUFSIZ] = { '\0' };
     Scm_libarchive_archive *p = NULL;
     struct archive *        ext = NULL;
     struct archive *        a = NULL;
     struct archive_entry *  entry_iter = NULL;

     p = SCM_LIBARCHIVE_ARCHIVE( archive );
     /* fprintf( stderr, "[%d:%d]\n", p->type, GAUCHE_ARCHIVE_READ ); */
     if( p->type != GAUCHE_ARCHIVE_READ ) {
          Scm_Error( "<archive-read> class required." );
     }

     flags = ARCHIVE_EXTRACT_TIME;
     flags |= ARCHIVE_EXTRACT_PERM;
     flags |= ARCHIVE_EXTRACT_ACL;
     flags |= ARCHIVE_EXTRACT_FFLAGS;

     a = archive_read_new();
     archive_read_support_filter_all( a );
     archive_read_support_format_all( a );
     ext = archive_write_disk_new();
     archive_write_disk_set_options( ext, flags );
     archive_write_disk_set_standard_lookup( ext );

     if( ( r = archive_read_open_filename(
               a, Scm_GetString( SCM_STRING( p->name ) ), BUFSIZ ) ) ) {
          Scm_Error( archive_error_string( a ) );
     }

     for( ;; ) {
          r = archive_read_next_header( a, &entry_iter );
          if( r == ARCHIVE_EOF ) {
               break;
          }
          if( r < ARCHIVE_OK ) {
               Scm_Error( archive_error_string( a ) );
          }
          if( r < ARCHIVE_WARN ) {
               Scm_Error( archive_error_string( a ) );
          }

          r = archive_write_header( ext, entry_iter );
          if( r < ARCHIVE_OK ) {
               Scm_Error( archive_error_string( ext ) );
          }
          if( archive_entry_size( entry_iter ) > 0 ) {
               len = archive_read_data( a, buf, sizeof( buf ) );
               while( len > 0 ) {
                    if( archive_write_data( ext, buf, len ) != len )
                         Scm_Error( archive_error_string( ext ) );
                    len = archive_read_data( a, buf, sizeof( buf ) );
               }
               if( len < 0 ) {
                    Scm_Error( archive_error_string( a ) );
               }
          }

          r = archive_write_finish_entry( ext );
          if( r < ARCHIVE_OK ) {
               Scm_Error( archive_error_string( ext ) );
          }
          if( r < ARCHIVE_WARN ) {
               Scm_Error( archive_error_string( ext ) );
          }
     }

     archive_read_close( a );
     archive_read_free( a );
     archive_write_close( ext );
     archive_write_free( ext );

     return SCM_TRUE;
}

ScmObj create_archive( ScmObj archive_name, ScmObj file_list,
                       ScmObj compress_format, ScmObj filter_format )
{
     struct archive *      a = NULL;
     struct archive_entry *entry = NULL;
     char                  buf[BUFSIZ] = { '\0' };
     int                   len = 0;
     int                   fd;
     const char *          s = NULL;
     ScmObj                p;
#ifdef HAVE_SYS_STAT_H
     struct stat st;
#endif /* #ifdef HAVE_SYS_STAT_H */

     a = archive_write_new();
     archive_write_set_format( a, Scm_GetInteger( compress_format ) );
     archive_write_add_filter( a, Scm_GetInteger( filter_format ) );
     archive_write_open_filename( a,
                                  Scm_GetString( SCM_STRING( archive_name ) ) );

     SCM_FOR_EACH( p, file_list )
     {
          s = Scm_GetString( SCM_STRING( SCM_CAR( p ) ) );
          entry = archive_entry_new();
          archive_entry_set_pathname( entry, s );
#ifdef HAVE_SYS_STAT_H
          stat( s, &st );
          archive_entry_set_size( entry, st.st_size );
          archive_entry_copy_stat( entry, &st );
          if( ( st.st_atim.tv_sec == 0 || st.st_ctim.tv_sec == 0 ) &&
              st.st_mtim.tv_sec != 0 ) {
               archive_entry_set_mtime( entry, st.st_mtim.tv_sec, 0 );
               archive_entry_set_atime( entry, st.st_mtim.tv_sec, 0 );
               archive_entry_set_ctime( entry, st.st_mtim.tv_sec, 0 );
          }
#endif /* #ifdef HAVE_SYS_STAT_H */
          archive_write_header( a, entry );
          fd = open( s, O_RDONLY );
          len = read( fd, buf, BUFSIZ );
          while( len > 0 ) {
               archive_write_data( a, buf, len );
               len = read( fd, buf, BUFSIZ );
          }
          close( fd );
          archive_entry_free( entry );
     }

     archive_write_close( a );
     archive_write_free( a );

     return SCM_TRUE;
}
